#!/usr/bin/env python
###########################################################################
#For the latest version and manual please visit the website (http://bix.ucsd.edu/projects/precioss/)
###########################################################################
# Title:           PreCIOSS.py
# Authors:          Roy Ronen, Glenn Tesler, Ali Akbari
# Created:         08/29/2015
# Last modified:   08/29/2015
# Description:     Apply PreCIOSS algorithm on SNP matrix
# Copyright (C) 2014-2015 The Regents of the University of California
# by Ali Akbari
#
# See file LICENSE for details.
###########################################################################
# This file is part of PreCIOSS.
#
# PreCIOSS is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, Version 2,
# dated June 1991, as published by the Free Software Foundation.
#
# PreCIOSS is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
###########################################################################

from numpy import *
import numpy as np
from sklearn import mixture
import sys
import matplotlib.pyplot as plt
###############################################################################
def convert_txt_to_matrix(address):
    f=open(address,'r')
    snp=f.read()
    snp=snp.splitlines()
    num_row=len(snp)
    num_col=len(snp[0])
    M=np.zeros((num_row,num_col))
    for i in range(num_row):
        M[i,:]=map(int,snp[i])
    f.close()
    return M 
###############################################################################
def haplotype_CFP_scores(hap_mat, freqs, subset_col=None, present=None, norm=2):
    ''' Computes CFP scores for all haplotypes (rows) in given matrix
        Returns np.array of scores (possibly containing np.nan)
        'hap_mat'   : haplotype matrix, numeric (0,1)
        'freqs'     : np.array of frequencies of corresponding matrix columns
        'subset_col': (optional) column based on which to subset results
        'present'   : (optional) boolean, indicates if column '1's should be kept '0''s
    '''
    cfp_exclude_fixed = True
    # initialize CFP scores vector
    scores = np.zeros(len(hap_mat))

    if cfp_exclude_fixed:
        # exclude fixed alleles, converting 1.0 to 0.0
        freqs_mod = np.copy(freqs)
        freqs_mod[freqs_mod == 1.0] = 0.0
    else:
        freqs_mod = freqs

    # de normalize, REMOVE??
    freqs_mod = freqs_mod * len(hap_mat)

    # compute CFP-score for each haplotype
    for (i, hap) in enumerate(hap_mat):

        scores[i] = np.nan  # exclude by defualt

        if subset_col is None:
            # no subset, include
            scores[i] = linalg.norm(np.multiply(hap, freqs_mod), norm)
        else:
            # subset, decide inclusion
            if present and hap_mat[i, subset_col] == 1.0:
                # include carrier i
                scores[i] = linalg.norm(np.multiply(hap, freqs_mod), norm)

            elif (not present) and hap_mat[i, subset_col] == 0.0:
                # include non-carrier i
                scores[i] = linalg.norm(np.multiply(hap, freqs_mod), norm)

    return scores

###############################################################################
def cluster_CFP_scores_GMM(scores):

    # prep
    scores = np.array(scores)
    X = np.reshape(np.array(scores), (-1, 1))
    X=X/max(X)
    threshold=median(X)-2*(std(X))
    I=np.nonzero(X>threshold)[0]
    # fit model
    g = mixture.GMM(n_components=2)
    g.fit(X[I])
    # predict labels
    labels = g.predict(X)
    high_gaussian_ind = np.argmax(g.means_[:, 0])
    low_gaussian_ind = np.argmin(g.means_[:, 0])
    pred_labels = np.array(labels == high_gaussian_ind, dtype=float)

    # class posteriors
    probs = g.predict_proba(X)
    p_carr = probs[:, high_gaussian_ind]
    p_ncarr = probs[:, low_gaussian_ind]

    return pred_labels, p_carr, p_ncarr
###############################################################################
def go(in_dir):
    out_dir='%s.report'%(in_dir)
    norm=1
    hap_mat=convert_txt_to_matrix(in_dir)
    col_freqs=np.mean(hap_mat,0)
    # 1. CFP scores
    hap_scores = haplotype_CFP_scores(hap_mat,col_freqs, norm=norm)
    # 2. cluster & predict
    pred_status, p_ycarr, p_ncarr= cluster_CFP_scores_GMM(hap_scores)
    # sanity check
    if len(hap_scores[pred_status == 0]) > 0 and len(hap_scores[pred_status == 1]) > 0:
        if not max(hap_scores[pred_status == 0]) <= min(hap_scores[pred_status == 1]):
            print "Warning#1: Variance of one cluster is too low!"
    else:
        print "Warning#2: GMM Could not find two separate clusters!"
    pval=pred_status*p_ncarr+(1-pred_status)*p_ycarr
    # 4. Report & save
    I=np.asarray(range(1,len(hap_mat)+1))
    report=np.asarray([I,hap_scores,pred_status,pval]).T
    np.savetxt(out_dir,report,fmt='%g')
    plt.hist([hap_scores[pred_status==0],hap_scores[pred_status==1]])
    plt.legend(['Non-Carriers','Carriers'],loc=0)
    plt.xlabel('HAF-score', size=20)
    plt.savefig('%s.pdf'%in_dir)
    return report

###############################################################################
#                                  MAIN                                       #
###############################################################################
if __name__ == '__main__':

    if len(sys.argv)!=2:
        print ("Invalid!")
    else:
        in_dir=sys.argv[1]
        go(in_dir)
        print("Done!")
